# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import urwid
from . import urwidpatches

from ..main import (aioloop, cfg, cmdmgr, srvapi, helpmgr)
from ..settings.defaults import DEFAULT_KEYMAP


from .keymap import KeyMap
keymap = KeyMap(callback=lambda cmd,widget: cmdmgr(cmd, on_error=log.error))
for args in DEFAULT_KEYMAP:
    if args['action'][0] == '<' and args['action'][-1] == '>':
        args['action'] = keymap.key(args['action'])
    keymap.bind(**args)
helpmgr.keymap = keymap


from .group import Group
from .tabs import Tabs
from .cli import CLIEditWidget
from .logger import LogWidget
from .infobar import (QuickHelpWidget, ConnectionStatusWidget,
                      BandwidthStatusWidget, TorrentCountersWidget)


def _create_cli_widget():
    def on_cancel(widget):
        widget.set_edit_text('')
        widgets.hide('cli')

    def on_accept(widget):
        cmd = widget.get_edit_text()
        on_cancel(widget)
        cmdmgr(cmd, on_error=log.error)

    return CLIEditWidget(':',
                         on_accept=on_accept, on_cancel=on_cancel,
                         history_file=cfg['tui.cli.history'].value)

def _greedy_spacer():
    return urwid.Padding(urwid.Text(''))

topbar = Group(cls=urwid.Columns)
topbar.add(name='host',   widget=ConnectionStatusWidget(), options='pack')
topbar.add(name='spacer', widget=urwid.AttrMap(_greedy_spacer(), 'topbar'))
topbar.add(name='help',   widget=QuickHelpWidget(), options='pack')

tabs = keymap.wrap(Tabs, context='tabs')(
    tabbar=urwid.AttrMap(urwid.Columns([], dividechars=2), 'tabs.unfocused')
)

bottombar = Group(cls=urwid.Columns)
bottombar.add(name='counters', widget=TorrentCountersWidget(), options='pack')
bottombar.add(name='spacer', widget=urwid.AttrMap(_greedy_spacer(), 'bottombar'))
bottombar.add(name='bandwidth', widget=BandwidthStatusWidget(), options='pack')

cli = urwid.AttrMap(_create_cli_widget(), 'cli')

logwidget = LogWidget(maxrows=cfg['tui.log.height'].value,
                      autohide_delay=cfg['tui.log.autohide'].value)

widgets = keymap.wrap(Group, context='main')(cls=urwid.Pile)
widgets.add(name='topbar', widget=topbar, options='pack')
widgets.add(name='main', widget=tabs)
widgets.add(name='log', widget=logwidget, options='pack')
widgets.add(name='cli', widget=cli, options='pack', visible=False)
widgets.add(name='bottombar', widget=bottombar, options='pack')


def unhandled_input(key):
    key = keymap.evaluate(key)
    if key is not None:
        log.debug('Unhandled key: %s', key)

urwidscreen = urwid.raw_display.Screen()
urwidloop = urwid.MainLoop(widgets,
                           screen=urwidscreen,
                           event_loop=urwid.AsyncioEventLoop(loop=aioloop),
                           unhandled_input=unhandled_input,
                           handle_mouse=False)


def run(cmds):
    """Run commands and start TUI

    Return False if any of the commands failed, True otherwise.
    """
    # Load tui-specific hooks before commands run (commands may trigger hooks)
    from . import hooks

    for cmd in cmds:
        try:
            if not cmdmgr(cmd, block=True, on_error=log.error):
                return False
        # Make 'quit' behave as expected
        except urwid.ExitMainLoop:
            return True

    # Don't catch theme.ParserError.  Default theme should throw exceptions to
    # stdout.
    from . import theme
    theme.init(cfg['tui.theme'].value, urwidscreen)

    # Enable logging to tui widget instead of stdout/stderr
    logwidget.enable()

    # If no tabs have been opened by cli or rc file, open default tab
    if len(tabs) <= 0:
        cmdmgr('ls', on_error=log.error)

    try:
        # Start polling torrent lists, counters, bandwidth usage, etc.
        aioloop.run_until_complete(srvapi.start_polling())
        urwidloop.run()
    finally:
        logwidget.disable()
        aioloop.run_until_complete(srvapi.stop_polling())

    return True



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Torrent commands for the TUI"""

from ...logging import make_logger
log = make_logger(__name__)

from collections import abc

from ..base import torrent as base
from . import mixin
from .. import ExpectedResource
from ...utils import strcrop
from . import make_tab_title


class AddTorrentsCmd(base.AddTorrentsCmdbase,
                     mixin.polling_frenzy, mixin.make_request):
    provides = {'tui'}


class ListTorrentsCmd(base.ListTorrentsCmdbase,
                      mixin.select_torrents):
    provides = {'tui'}
    tui = ExpectedResource

    def make_tlist(self, tfilter, sort, columns):
        from ...tui.torrent.tlist import TorrentListWidget
        tlistw = TorrentListWidget(tfilter=tfilter, sort=sort, columns=columns)
        titlew = make_tab_title(tlistw.title, 'tabs.torrentlist.unfocused', 'tabs.torrentlist.focused')
        self.tui.tabs.load(titlew, tlistw)
        return True


class ListFilesCmd(base.ListFilesCmdbase,
                   mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}
    tui = ExpectedResource
    srvapi = ExpectedResource

    async def make_flist(self, tfilter, ffilter, columns):
        from ...tui.torrent.flist import FileListWidget
        flistw = FileListWidget(self.srvapi, tfilter, ffilter, columns)

        if isinstance(tfilter, abc.Sequence) and len(tfilter) == 1:
            # tfilter is a torrent ID - resolve it to a name for the title
            response = await self.srvapi.torrent.torrents(tfilter, keys=('name',))
            title = strcrop(response.torrents[0]['name'], 30, tail='…')
        else:
            if ffilter is None:
                title = str(tfilter)
            else:
                title = '%s files of %s torrents' % (ffilter, tfilter)
        titlew = make_tab_title(title, 'tabs.filelist.unfocused', 'tabs.filelist.focused')
        self.tui.tabs.load(titlew, flistw)
        return True


class PriorityCmd(base.PriorityCmdbase,
                  mixin.polling_frenzy, mixin.make_request, mixin.select_torrents, mixin.select_files):
    provides = {'tui'}


class RemoveTorrentsCmd(base.RemoveTorrentsCmdbase,
                        mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StartTorrentsCmd(base.StartTorrentsCmdbase,
                       mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class StopTorrentsCmd(base.StopTorrentsCmdbase,
                      mixin.polling_frenzy, mixin.make_request, mixin.select_torrents):
    provides = {'tui'}


class VerifyTorrentsCmd(base.VerifyTorrentsCmdbase,
                        mixin.make_request, mixin.select_torrents):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Commands that work exclusively in the TUI"""

from ...logging import make_logger
log = make_logger(__name__)


from .. import (InitCommand, ExpectedResource)


from ...tui import KEYMAP_CONTEXTS
class BindCmd(metaclass=InitCommand):
    name = 'bind'
    provides = {'tui'}
    category = 'tui'
    description = 'Bind keys to commands or other keys'
    usage = ('bind [--context <CONTEXT>] <KEY> <COMMAND>',)
    examples = ('bind --context tabs alt-[ tab --focus left',
                'bind --context tabs alt-] tab --focus right',
                'bind --context torrent alt-! start --force',
                'bind ctrl-a tab ls active',
                'bind u <up>',
                'bind d <down>')
    argspecs = (
        { 'names': ('--context','-c'),
          'description': 'Where KEY is grabbed (see CONTEXTS section)' },
        { 'names': ('KEY',),
          'description': 'Key or key combination (see KEYS section)' },
        { 'names': ('COMMAND',), 'nargs': 'REMAINDER',
          'description': ("Any command or '<KEY>' (including the brackets) "
                          'to translate one key to another') },
    )

    def __create_CONTEXTS_section():
        lines = [
            ('Each key is mapped in a specific context or as a default mapping '
             'that is used when no context wants the key.  The same key can be '
             'mapped to different actions in different contexts.'),
            '',
            'Available contexts: ' + \
              ', '.join("'%s'" % context for context in KEYMAP_CONTEXTS),
            '',
            'EXAMPLE',
            '\tbind --context torrent ctrl-t start',
            '\tbind --context tabs ctrl-t tab',
            '\tbind ctrl-t <left>',
            '',
            ("\tWhen focusing a torrent, 'ctrl-t' starts the focused torrent.  "
             "If focus is not on a torrent but still on a tab (e.g. when reading "
             "documentation) a new tab is opened.  Otherwise (e.g. focus is on the "
             "command prompt), 'ctrl-t' does the same as 'left'."),
        ]
        return lines

    more_sections = {
        'CONTEXTS': __create_CONTEXTS_section,
        'KEYS': (
            ("Single-character keys are specified by typing them (e.g. 'h', "
             "'X', '5', '!', 'þ', etc).  Special key names are 'enter', 'space', "
             "'tab', 'backspace', 'insert', 'delete', 'home', 'end', 'down', 'up', "
             "'left', 'right', 'pgup', 'pgdn' and 'f<1-12>'."),
            '',
            ("The modifiers 'ctrl', 'alt' and 'shift' are combined with '-' "
             "or ' ', e.g. 'alt-i', 'shift-delete', 'ctrl a', etc.  "
             "The 'shift' modifier is only recognized for "
             "special keys like 'delete'; 'shift-t' does not work."),
        ),
    }

    tui = ExpectedResource

    def run(self, context, KEY, COMMAND):
        keymap = self.tui.keymap
        if len(COMMAND) == 1 and \
           COMMAND[0][0] == '<' and COMMAND[0][-1] == '>':
            # COMMAND is another key (e.g. 'j' -> 'down')
            COMMAND = keymap.key(COMMAND[0])

        if context is not None and context not in KEYMAP_CONTEXTS:
            log.error('Invalid context: {!r}'.format(context))
            return False
        try:
            keymap.bind(KEY, COMMAND, context=context)
        except ValueError as e:
            log.error(e)
            return False
        else:
            return True


class ClearLogCmd(metaclass=InitCommand):
    name = 'clearlog'
    provides = {'tui'}
    category = 'tui'
    description = 'Clear all logged messages'

    tui = ExpectedResource

    def run(self):
        logwidget = self.tui.logwidget
        if len(tuple(logwidget.entries)) > 0:
            logwidget.clear()
            return True
        else:
            return False


class QuitCmd(metaclass=InitCommand):
    name = 'quit'
    provides = {'tui'}
    category = 'tui'
    description = 'Terminate the TUI'

    def run(self):
        import urwid
        raise urwid.ExitMainLoop()


class RcCmd(metaclass=InitCommand):
    name = 'rc'
    aliases = ('source',)
    category = 'misc'
    provides = {'tui'}
    description = 'Run commands in rc file'
    usage = ('rc <FILE>',)
    examples = ('rc rc.example.org   # Load $XDG_CONFIG_HOME/.config/{APPNAME}/rc.example.org',)
    argspecs = (
        {'names': ('FILE',),
         'description': ('Path to rc file; if FILE does not start with '
                         "'.', '~' or '/', $XDG_CONFIG_HOME/.config/{APPNAME}/ "
                         'is prepended')},
    )
    cmdmgr = ExpectedResource

    def run(self, FILE):
        from ...settings import rcfile
        from ...settings.defaults import DEFAULT_RCFILE
        import os

        if FILE[0] not in (os.sep, '.', '~'):
            FILE = '{}{}{}'.format(os.path.dirname(DEFAULT_RCFILE),
                                   os.sep,
                                   FILE)

        try:
            lines = rcfile.read(FILE)
        except rcfile.RcFileError as e:
            log.error('Loading rc file failed: {}'.format(e))
            return False
        else:
            log.debug('Reading commands from rc file: %r', FILE)
            log.debug(lines)
            for cmdline in lines:
                self.cmdmgr(cmdline)
            return True


class TabCmd(metaclass=InitCommand):
    name = 'tab'
    provides = {'tui'}
    category = 'tui'
    description = 'Open, close and focus tabs'
    usage = ('tab [--close] [--focus <TITLE OR INDEX>] [<COMMAND>]',)
    examples = ('tab',
                'tab ls active',
                'tab -f active',
                'tab -f 3 ls active',
                'tab -c')
    argspecs = (
        { 'names': ('--close', '-c'), 'nargs': '?', 'const': None, 'default': False,
          'description': 'Close tab at index CLOSE or with partial tital CLOSE' },
        { 'names': ('--focus', '-f'),
          'description': ('Focus tab; FOCUS can be an index (first tab is 1), '
                          "part of a tab title, 'left'/'-' or 'right'/'+'") },
        { 'names': ('COMMAND',), 'nargs': 'REMAINDER',
          'description': ('Command to run in new tab') },
    )

    tui = ExpectedResource
    cmdmgr = ExpectedResource

    async def run(self, close, focus, COMMAND):
        import urwid
        tabs = self.tui.tabs
        cmdargs = {}

        if close is not False:
            log.debug('Closing tab %r', close)
            i = self.get_tab_index(close, tabs)
            if i is not None:
                tabs.remove(i)
            else:
                return False

        if focus is not None:
            log.debug('Focusing tab %r', focus)
            i = self.get_tab_index(focus, tabs)
            if i is not None:
                tabs.focus_position = i
            else:
                return False

        # Remember which torrent is focused in current tab so we can provide
        # it to the command running in a new tab.
        try:
            cmdargs['focused_torrent'] = tabs.focus.focused_torrent.torrent
        except AttributeError:
            pass

        if close is False and focus is None:
            titlew = urwid.AttrMap(urwid.Text('Empty tab'), 'tabs.unfocused', 'tabs.focused')
            tabs.insert(titlew, position='right')
            log.debug('Inserted new tab at position %d', tabs.focus_position)

        if COMMAND:
            log.debug('Running command in tab %d with args %s: %r',
                      tabs.focus_position,
                      ', '.join('%s=%r' % (k,v) for k,v in cmdargs.items()),
                      COMMAND)

            process = self.cmdmgr.run(COMMAND, **cmdargs)
            # Sync processes are always finished at this point.
            # Async processes must finish before we can report our own success.
            if not process.finished:
                await process.task
            return process.success
        else:
            return True

    @staticmethod
    def get_tab_index(pos, tabs):
        if pos is None:
            return tabs.focus_position

        def find_index(i):
            try:
                return tabs.get_index(i-1)
            except IndexError as e:
                log.error('No tab at position: {}'.format(e.value+1))

        def find_title(string):
            for i,title in enumerate(tabs.titles):
                if string in title.original_widget.text:
                    return i
            log.error('No tab found: {!r}'.format(string))

        tabcount = len(tabs)
        curpos = tabs.focus_position
        curpos = 1 if curpos is None else curpos

        if pos.isdigit():
            return find_index(int(pos))
        elif pos in ('left', '-'):
            if tabcount > 1:
                return tabs.get_index(max(0, curpos-1))
        elif pos in ('right', '+'):
            if tabcount > 1:
                return tabs.get_index(min(tabcount-1, curpos+1))
        else:
            return find_title(pos)


class TUICmd(metaclass=InitCommand):
    name = 'tui'
    provides = {'tui'}
    category = 'tui'
    description = 'Show or hide parts of the text user interface'
    usage = ('tui <ACTION> <ELEMENT> <ELEMENT> ...',)
    examples = ('tui toggle log',
                'tui hide topbar.help')
    argspecs = (
        { 'names': ('ACTION',), 'choices': ('show', 'hide', 'toggle'),
          'description': '"show", "hide" or "toggle"' },
        { 'names': ('ELEMENTS',), 'nargs': '+',
          'description': ('Name(s) of TUI elements; '
                          'see ELEMENT NAMES section for a list') },
    )

    # Lazily load the element names from tui (HelpManager supports sequences
    # of lines or a callable that returns them)
    def __create_element_names():
        from ...tui.main import widgets
        return ('Available TUI element names are: ' + \
                ', '.join(str(e) for e in sorted(widgets.names_recursive)),)
    more_sections = { 'ELEMENT NAMES': __create_element_names }

    tui = ExpectedResource

    def run(self, ACTION, ELEMENTS):
        widgets = self.tui.widgets
        widget = None
        success = False
        for element in ELEMENTS:
            # Resolve path
            path = element.split('.')
            target_name = path.pop(-1)
            current_path = []
            widget = widgets
            try:
                for widgetname in path:
                    current_path.append(widgetname)
                    widget = getattr(widget, widgetname)
            except AttributeError as e:
                log.error('Unknown TUI element: %r', '.'.join(current_path))

        if widget is not None:
            action = getattr(widget, ACTION)
            log.debug('%sing %s in %s', ACTION.capitalize(), target_name, widget)
            try:
                action(target_name)
                success = True
            except ValueError as e:
                log.error(e)
        return success



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from functools import partial


class _Sorter():
    def __init__(self, *keyfuncs, needed_keys, description, aliases=()):
        self._keyfuncs = keyfuncs
        self.needed_keys = needed_keys
        self.description = 'Sort torrents by ' + description
        self.aliases = aliases

    def __call__(self, items, reverse=False, inplace=False, torrent_getter=lambda item: item):
        if not items:
            return items

        for keyfunc in self._keyfuncs:
            def key_getter(item):
                return keyfunc(torrent_getter(item))

            if inplace:
                items[:] = sorted(items, key=key_getter, reverse=reverse)
            else:
                items = sorted(items, key=key_getter, reverse=reverse)

        return items


SORTERS = {
    'name':         _Sorter(lambda t: t['name'].lower(),
                            needed_keys=('name',),
                            description='name'),
    'path':         _Sorter(lambda t: t['path'],
                            aliases=('dir',),
                            needed_keys=('path',),
                            description='download location'),
    'status':       _Sorter(lambda t: t['status'],
                            needed_keys=('status',),
                            description='status (downloading, seeding, verifying, etc.)'),
    'size':         _Sorter(lambda t: t['size-final'],
                            needed_keys=('size-final',),
                            description='size'),
    'peers':        _Sorter(lambda t: t['peers-connected'],
                            needed_keys=('peers-connected',),
                            description='connected peers'),
    'seeds':        _Sorter(lambda t: t['peers-seeding'],
                            needed_keys=('peers-seeding',),
                            description='domain of first tracker'),
    'ratio':        _Sorter(lambda t: t['ratio'],
                            needed_keys=('ratio',),
                            description='ratio'),
    'rate-down':    _Sorter(lambda t: t['rate-down'],
                            needed_keys=('rate-down',),
                            description='download rate'),
    'rate-up':      _Sorter(lambda t: t['rate-up'],
                            needed_keys=('rate-up',),
                            description='upload rate'),
    'rate':         _Sorter(lambda t: t['rate-up'] + t['rate-down'],
                            needed_keys=('rate-up', 'rate-down'),
                            description='combined download and upload rate'),
    'uploaded':     _Sorter(lambda t: t['size-uploaded'],
                            needed_keys=('size-uploaded',),
                            description='number of uploaded bytes'),
    'downloaded':   _Sorter(lambda t: t['size-downloaded'],
                            needed_keys=('size-downloaded',),
                            description='number of downloaded bytes'),
    'progress':     _Sorter(lambda t: t['%downloaded'],
                            lambda t: t['%metadata'],
                            lambda t: t['%verified'],
                            needed_keys=('%downloaded', '%metadata', '%verified'),
                            description='downloading or verifying progress'),
    'tracker':      _Sorter(lambda t: t['trackers'][0]['url-announce'].domain if t['trackers'] else '',
                            needed_keys=('trackers',),
                            description='domain of first tracker'),
}

# Map aliases to their original name
_ALIASES = {alias: sname
            for sname,s in SORTERS.items()
            for alias in s.aliases}


class TorrentSorter():
    def __init__(self, sortstrings=()):
        sorters = []         # List of (<_Sorter instance>, <reverse=True/False>) tuples
        needed_keys = set()  # Torrent keys needed by all sorters
        strings = []         # String reprs of sorters
        for sortstring in sortstrings:
            if sortstring.startswith('!'):
                sortername, reverse = sortstring[1:], True
            else:
                sortername, reverse = sortstring, False

            # Resolve alias
            if sortername in _ALIASES:
                sortername = _ALIASES[sortername]

            if sortername not in SORTERS:
                raise ValueError('Unknown sort order: {}'.format(sortername))
            else:
                sorter = SORTERS[sortername]
                sorters.append(partial(sorter, reverse=reverse))
                strings.append(('!' if reverse else '') + sortername)
                needed_keys.update(sorter.needed_keys)

        # Insert alphabetical sorter unless one already exists (reverse or not)
        if 'name' not in sortstrings:
            namesorter = SORTERS['name']
            sorters.insert(0, namesorter)
            needed_keys.update(namesorter.needed_keys)

        self._sorters = tuple(sorters)
        self._str = ','.join(strings)
        self._needed_keys = tuple(needed_keys)

    def apply(self, items, inplace=False, torrent_getter=lambda item: item):
        """Sort sequence `items` in-place or return new list of items

        torrent_getter: Callable that gets an item of `items` and returns a
                        Torrent object (e.g. to sort lists of torrent widgets
                        in-place)
        inplace: Modify `items` if True, otherwise return a new list
        """
        import time
        start_time = time.time()

        for sorter in self._sorters:
            items = sorter(items, inplace=inplace, torrent_getter=torrent_getter)

        log.debug('-> Sorted %d items by %s in %.3fms',
                  len(items), self, (time.time()-start_time)*1e3)

        if not inplace:
            return items

    @property
    def needed_keys(self):
        return self._needed_keys

    def __str__(self):
        return self._str

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, str(self))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..main import (cfg, srvapi, aioloop)
from . import main as tui
from ..client import ClientError


def _connect_to_new_url(url):
    # See also ..hooks
    async def coro():
        try:
            await srvapi.rpc.connect(url.value)
        except ClientError as e:
            pass
    aioloop.create_task(coro())
cfg['srv.url'].on_change(_connect_to_new_url)

def _update_pollers(seconds):
    tui.srvapi.poll()
srvapi.rpc.on('connected', _update_pollers)

def _clear_list_widgets(seconds):
    from .torrent.tlist import TorrentListWidget
    from .torrent.flist import FileListWidget
    for widget in tui.tabs.contents:
        if isinstance(widget, (TorrentListWidget, FileListWidget)):
            widget.clear()
srvapi.rpc.on('disconnected', _clear_list_widgets)

def _set_poll_interval(seconds):
    tui.srvapi.interval = seconds.value
cfg['tui.poll'].on_change(_set_poll_interval)

def _set_cli_history(histfile):
    tui.cli.original_widget.history_file = histfile.value
cfg['tui.cli.history'].on_change(_set_cli_history)

def _set_autohide_delay(seconds):
    tui.log.autohide_delay = seconds.value
cfg['tui.log.autohide'].on_change(_set_autohide_delay)

def _set_log_maxrows(rows):
    tui.log.maxrows = rows.value
cfg['tui.log.height'].on_change(_set_log_maxrows)

def _set_theme(themefile):
    try:
        tui.theme.load(cfg['tui.theme'].value, tui.urwidscreen)
    except tui.theme.ThemeError as e:
        raise ValueError(e)
cfg['tui.theme'].on_change(_set_theme)

def _refresh_tlists(value):
    from .torrent.tlist import TorrentListWidget
    from .torrent.flist import FileListWidget
    for widget in tui.tabs:
        if isinstance(widget, (TorrentListWidget, FileListWidget)):
            widget.clear()
    tui.srvapi.poll()
cfg['unit.bandwidth'].on_change(_refresh_tlists)
cfg['unit.size'].on_change(_refresh_tlists)
cfg['unitprefix.bandwidth'].on_change(_refresh_tlists)
cfg['unitprefix.size'].on_change(_refresh_tlists)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column specifications for tables"""

class ColumnBase():
    header = {'left': '', 'right': ''}
    width = None
    align = 'right'

    def __init__(self, data=None):
        self.data = data if data is not None else {}
        super().__init__()

    def get_value(self):
        raise NotImplementedError()

    def get_raw(self):
        return self.get_value()

    def get_string(self):
        """Return `get_value` as spaced and aligned string

        If the `width` attribute is not set to None, expand or shrink and
        align the returned string (`align` attribute must be 'left' or
        'right').
        """
        text = str(self.get_value())
        if not isinstance(self.width, int):
            return text
        else:
            return self._crop_and_align(text, self.width, self.align)

    def _crop_and_align(self, string, width, side):
        if len(string) > width:
            string = string[:width]

        if side == 'right':
            return string.rjust(width)
        elif side == 'left':
            return string.ljust(width)
        else:
            raise TypeError("side argument must be 'left' or 'right', not {!r}".format(side))

    def __repr__(self):
        if self.data:
            return '<{} {}>'.format(type(self).__name__, self.get_value())
        else:
            return '<{} <UNINITIALIZED>>'.format(type(self).__name__)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""General hooks that are always needed regardless of the interface"""

from .main import (cfg, srvapi)
from .columns.tlist import COLUMNS as TCOLUMNS
from .columns.flist import COLUMNS as FCOLUMNS


def _set_rpc_timeout(timeout):
    srvapi.rpc.timeout = timeout.value
cfg['srv.timeout'].on_change(_set_rpc_timeout)


def _set_rpc_url(url):
    srvapi.url = url.value
cfg['srv.url'].on_change(_set_rpc_url)


def _set_bandwidth_unit(unit):
    srvapi.bandwidth_unit = unit.value
    u = {'bit': 'b', 'byte': 'B'}[unit.value]
    TCOLUMNS['rate-up'].set_unit(u)
    TCOLUMNS['rate-down'].set_unit(u)
cfg['unit.bandwidth'].on_change(_set_bandwidth_unit)
_set_bandwidth_unit(cfg['unit.bandwidth'])  # Initially call TCOLUMNS[...].set_unit()


def _set_bandwidth_prefix(prefix):
    srvapi.bandwidth_prefix = prefix.value
cfg['unitprefix.bandwidth'].on_change(_set_bandwidth_prefix)


def _set_size_unit(unit):
    srvapi.size_unit = unit.value
    u = {'bit': 'b', 'byte': 'B'}[unit.value]
    TCOLUMNS['size'].set_unit(u)
    TCOLUMNS['downloaded'].set_unit(u)
    TCOLUMNS['uploaded'].set_unit(u)

    FCOLUMNS['size'].set_unit(u)
    FCOLUMNS['downloaded'].set_unit(u)
cfg['unit.size'].on_change(_set_size_unit)
_set_size_unit(cfg['unit.size'])  # Initially call TCOLUMNS[...].set_unit()


def _set_size_prefix(prefix):
    srvapi.size_prefix = prefix.value
cfg['unitprefix.size'].on_change(_set_size_prefix)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import argparse
from collections import OrderedDict


parser = argparse.ArgumentParser(add_help=False)
descriptions = OrderedDict()
def add_arg(*args, section='OPTIONS', description=None, varname=None, **kwargs):
    if description is not None:
        optstr = ','.join(args)
        if varname is not None:
            optstr += ' %s' % varname

        if section not in descriptions:
            descriptions[section] = OrderedDict()
        descriptions[section][optstr] = description
    parser.add_argument(*args, **kwargs)


add_arg('--help','-h', nargs='*', default=None,
        section='OPTIONS',
        description="Display help about TOPIC",
        varname='[TOPIC]')

add_arg('--version','-v', action='store_true',
        section='OPTIONS',
        description="Display version number and exit")

add_arg('--tui', '-t', action='store_true',
        section='OPTIONS',
        description='Enforce the TUI')
add_arg('--notui', '--no-tui', '-T', action='store_true',
        section='OPTIONS',
        description='Inhibit the TUI')

add_arg('--rcfile', '--rc-file', '-c',
        section='OPTIONS',
        description='Run commands from FILE upon startup',
        varname='FILE')
add_arg('--norcfile', '--no-rc-file', '-C', action='store_true',
        section='OPTIONS',
        description='Do not run commands from any rc file')

add_arg('--debug', type=lambda mods: mods.split(','), default=[],
        section='DEVELOPER OPTIONS',
        description=('Log debug messages from comma-separated list of MODULES'
                     ' (e.g. "client,commands.tui")'),
        varname='MODULES')
add_arg('--debug-file', default=None,
        section='DEVELOPER OPTIONS',
        description='Log debug messages to FILE',
        varname='FILE')
add_arg('--profile-file', default=None,
        section='DEVELOPER OPTIONS',
        description='Write cProfile statistics to FILE',
        varname='FILE')

# Anything not specified above is a subcommand or a subcommand option.
parser.add_argument('subcmds', nargs=argparse.REMAINDER)
ARGS = vars(parser.parse_args())
_SUBCMDS = ARGS.pop('subcmds')

# Convert -h option to 'help' command
if ARGS['help'] is not None:
    _SUBCMDS.append('help')
    _SUBCMDS.extend(ARGS['help'])

# Convert -v option to 'version' command
if ARGS['version']:
    _SUBCMDS.append('version')


_CMD_SEPARATORS = (';',)

def get_cmds():
    """Return tuple of commands, each command being a tuple of arguments"""
    cmds = [[]]
    for arg in _SUBCMDS:
        if arg in _CMD_SEPARATORS:
            cmds.append([])       # Start new command
        else:
            cmds[-1].append(arg)  # Append arg to current command
    return tuple(tuple(cmd) for cmd in cmds if cmd)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import re


class Key(str):
    """Key code as string

    This class prettifies urwid key codes and accepts some alternatives, for
    example Key('del') creates Key('delete'), Key('escape') -> Key('esc'),
    etc.
    """

    _PARSE_MAP = (
        (re.compile(r'^<(.+)>$'),                 r'\1'),
        (re.compile(r'(.+)-(.+)'),                r'\1 \2'),
        (re.compile(r'^alt', flags=re.I),         r'meta'),
        (re.compile(r'^space$', flags=re.I),      r' '),
        (re.compile(r'^return$', flags=re.I),     r'enter'),
        (re.compile(r'^escape$', flags=re.I),     r'esc'),
        (re.compile(r'^pgup$', flags=re.I),       r'page up'),
        (re.compile(r'^pgdn$', flags=re.I),       r'page down'),
        (re.compile(r'^pos1$', flags=re.I),       r'home'),
        (re.compile(r'^ins$', flags=re.I),        r'insert'),
        (re.compile(r'^del$', flags=re.I),        r'delete'),
        (re.compile(r'^(\w+) (\S+)$'),
         lambda m: m.group(1).lower()+' '+m.group(2)),
    )

    def __new__(cls, key):
        # Conform to urwid conventions
        for regex,repl in cls._PARSE_MAP:
            key = regex.sub(repl, key)
        return super().__new__(cls, key)

    _PRETTIFY_MAP = (
        (re.compile(r'^esc$'), r'escape'),
        (re.compile(r'^ $'),   r'space'),
        (re.compile(r'^meta'), r'alt'),
        (re.compile(r' '),     r'-'),
    )

    @property
    def pretty(self):
        """Pretty string representation"""
        key = self
        for regex,repl in self._PRETTIFY_MAP:
            key = regex.sub(repl, key)
        return key

    def __repr__(self):
        return '<'+str(self)+'>'


class KeyMapped():
    """Mixin class that provides a keypress method"""
    def keypress(self, size, key):
        sup = super()
        def try_super(key):
            if sup.selectable():
                key = sup.keypress(size, key)
            return key

        # Offer key to parent's keypress()
        key = try_super(key)

        # If parent doesn't want key, check if it is mapped
        if key is not None:
            key = self.keymap.evaluate(key,
                                       context=self.context,
                                       callback=self.callback,
                                       widget=self)

            # evaluate() may have resolved key to another key
            # (e.g. 'j' -> 'down')
            if key is not None:
                key = try_super(key)

        return key

    def selectable(self):
        return True


class KeyMap():
    """Bind keys to actions in different contexts

    To bind keys to widgets, a new class is created that accepts user input by
    adding the methods `keypress` and `selectable` to the original class:

    >>> keymap = KeyMap()
    >>> keymap.bind(key='meta g',
                    action=lambda widget: widget.set_edit_text('abcfoo'),
                    context='pwentry')
    >>> Password = keymap.wrap(urwid.Edit, context='pwentry')
    >>> Password
    urwid.Edit_KeyMapped

    Now you can use Password like any regular Edit widget:

    >>> pw = Password('', 'Enter password or <Alt-g> to generate one')
    """

    def __init__(self, callback=None):
        self._callback = callback
        self._contexts = {None: {}}
        self._clscache = {}

    def bind(self, key, action, context=None):
        """Bind `key` to `action` in `context`

        key: an urwid key string
        context: a descriptive string (or any other hashable object)
        action: What to do when `key` is pressed in `context`; may be:
                - a callable that accepts the widget that received `key`,
                - a Key object (so you can bind <ctrl a> to whatever <ctrl b>
                  is bound to,
                - any other object that is passed to the `callback` specified
                  in calls to `wrap` or when the KeyMap was instantiated.
        """
        key = Key(key)
        if isinstance(action, Key) and key == action:
            raise ValueError('Mapping {!r} to {!r} is silly'.format(key, action))

        if context not in self._contexts:
            self._contexts[context] = {}
        self._contexts[context][key] = action
        log.debug('%s: Mapped %r -> %r', context, key, action)

    def unbind(self, key, context=None):
        """Unbind `key` in `context`"""
        key = Key(key)
        if context not in self._contexts:
            raise ValueError('Unknown context: {!r}'.format(context))
        elif key not in self._contexts[context]:
            raise ValueError('Key not mapped in context {!r}: {!r}'.format(context, key))
        else:
            del self._contexts[context][key]
            log.debug('Unmapped %r [%s]', key, context)

    def evaluate(self, key, context=None, callback=None, widget=None):
        """Run action that is bound to `key` in `context`

        key: the pressed key
        context: the context `key` was pressed in
        callback: None or any callable
        widget: the widget that received `key`
        """
        if context not in self._contexts:
            raise ValueError('Unknown context: {}'.format(context))

        # Find the action that is bound to key in context
        key = Key(key)
        if key in self._contexts[context]:
            action = self._contexts[context][key]
            log.debug('Evaluated %r in context %r: %r', key, context, action)
        elif key in self._contexts[None]:
            action = self._contexts[None][key]
            log.debug('Evaluated %r in default context: %r', key, action)
        else:
            log.debug('%r is not mapped in context %r', key, context)
            return key

        # Run the action
        if isinstance(action, Key):
            log.debug('Evaluating %r', action)
            return self.evaluate(action, context, callback, widget)
        elif callable(action):
            # action itself is the callback
            log.debug('Calling action: %r', action)
            action(widget)
        elif callback is not None:
            # Individual callback for this widget
            log.debug('Calling callback %r with widget %r', callback, widget)
            callback(action, widget)
        elif self._callback is not None:
            # General callback for all widgets
            log.debug('Calling callback %r with widget %r', self._callback, widget)
            self._callback(action, widget)
        else:
            raise TypeError('No callback given - unable to handle {!r}'.format(action))

    def wrap(self, cls, callback=None, context=None):
        """Return widget class that passes keypress()es through evaluate()"""
        clsid = id(cls)
        if clsid in self._clscache:
            cls_km = self._clscache[clsid]
        else:
            if context not in self._contexts:
                self._contexts[context] = {}

            log.debug('Wrapping %r in %r', cls, KeyMapped)
            clsname = cls.__name__ + '_KeyMapped'
            cls_km = type(clsname,
                          (KeyMapped, cls),
                          {'keymap': self,
                           'context': context,
                           'callback': staticmethod(callback)})
            self._clscache[clsid] = cls_km
        return cls_km

    def key(self, string):
        """Convenience method to make Key object from string"""
        return Key(string)

    def keys(self, action, context=None):
        """Yield keys that are mapped to actions that start with `action`

        If `context` is not None, yield only keys mapped in that context.
        """
        if context is not None:
            for k,a in self._contexts[context].items():
                if a.startswith(action):
                    yield k
        else:
            for context in self.contexts:
                for k,a in self._contexts[context].items():
                    if a.startswith(action):
                        yield k

    def map(self, context=None):
        """Yield (key, action) tuples for keys mapped in `context`"""
        yield from self._contexts[context].items()

    @property
    def contexts(self):
        """Yield existing contexts"""
        yield from self._contexts



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


class Constant():
    def __repr__(self):
        return '<Constant: {}>'.format(str(self).upper())

class _Disconnected(Constant):
    def __str__(self): return 'disconnected'
DISCONNECTED = _Disconnected()

class _Unlimited(Constant, float):
    def __str__(self): return 'unlimited'
UNLIMITED = _Unlimited('inf')  # float accepts 'inf'

class _Disabled(Constant):
    def __str__(self): return 'disabled'
    def __bool__(self): return False
DISABLED = _Disabled()

class _Enabled(Constant):
    def __str__(self): return 'enabled'
    def __bool__(self): return True
ENABLED = _Enabled()



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes that are common between TUI and CLI commands"""

from ...logging import make_logger
log = make_logger(__name__)


from .. import utils


class get_torrent_sorter():
    def get_torrent_sorter(self, args):
        """Return TorrentSorter instance or None

        If `args` evaluates to True, it is passed to TorrentSorter and the
        result is returned.

        If `args` evaluates to False, None is returned.
        """
        if args:
            from ...client import TorrentSorter
            return TorrentSorter(utils.listify_args(args))


class get_tlist_columns():
    def get_tlist_columns(self, args):
        """Check if each item in iterable `args` is a valid torrent list column

        Raise ValueError or return the same iterable.
        """
        from ...columns.tlist import COLUMNS
        args = utils.listify_args(args)
        for arg in args:
            if arg not in COLUMNS:
                raise ValueError('Unknown column name: {}'.format(arg))
        return args


class get_flist_columns():
    def get_flist_columns(self, args):
        """Check if each item in iterable `args` is a valid file list column

        Raise ValueError or return the same iterable.
        """
        from ...columns.flist import COLUMNS
        args = utils.listify_args(args)
        for arg in args:
            if arg not in COLUMNS:
                raise ValueError('Unknown column name: {}'.format(arg))
        return args



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import os
from appdirs import (user_cache_dir, user_config_dir)

from .. import APPNAME
from ..columns.tlist import COLUMNS as TCOLUMNS
from ..columns.flist import COLUMNS as FCOLUMNS
from ..client.tsort import SORTERS

DEFAULT_RCFILE        = user_config_dir(APPNAME) + '/rc'
DEFAULT_HISTORY_FILE  = user_cache_dir(APPNAME)+'/history'
DEFAULT_THEME_FILE    = os.path.dirname(__file__) + os.sep + 'default.theme'

DEFAULT_TLIST_SORT    = ('name',)
DEFAULT_TLIST_COLUMNS = ('name', 'ratio', 'size', 'downloaded', 'uploaded',
                         'eta', 'peers-connected', 'peers-seeding', 'rate-down',
                         'rate-up')
DEFAULT_FLIST_COLUMNS = ('priority', 'name', 'progress', 'downloaded', 'size')


from .settings import (StringValue, IntegerValue, NumberValue, BooleanValue,
                       PathValue, ListValue, SetValue, OptionValue)

class TorrentSortValue(SetValue):
    """SetValue that correctly validates inverted sort orders (e.g. '!name')"""
    def __init__(self, *args, options=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = tuple(SORTERS)

    def validate(self, names):
        for name in names:
            if name[0] == '!':
                name = name[1:]
            if name not in SORTERS:
                raise ValueError('Invalid sort order: {!r}'.format(name))


def init_defaults(cfg):
    cfg.load(
        StringValue('srv.url', default='http://localhost:9091',
                    description='URL of the Transmission RPC interface'),
        NumberValue('srv.timeout', default=10, min=0,
                    description=('Number of seconds before connecting '
                                 'to Transmission daemon fails')),

        SetValue('tlist.columns', default=DEFAULT_TLIST_COLUMNS,
                 options=tuple(TCOLUMNS),
                 description='List of columns in new torrent lists'),
        TorrentSortValue('tlist.sort', default=DEFAULT_TLIST_SORT,
                 description='List of torrent sort orders'),

        SetValue('flist.columns', default=DEFAULT_FLIST_COLUMNS,
                 options=tuple(FCOLUMNS),
                 description='List of columns in new torrent file lists'),


        PathValue('tui.theme', default=DEFAULT_THEME_FILE,
                  description='Path to theme file'),
        IntegerValue('tui.log.height', default=10, min=1,
                     description='Maximum height of the log section'),
        NumberValue('tui.log.autohide', default=10, min=0,
                    description=('If the log is hidden, show it for this many seconds '
                                 'for new log entries before hiding it again')),
        PathValue('tui.cli.history', default=DEFAULT_HISTORY_FILE,
                  description='Path to TUI command line history file'),
        NumberValue('tui.poll', default=5, min=0.1,
                    description='Interval in seconds between TUI updates'),

        OptionValue('unit.bandwidth', default='byte', options=('bit', 'byte'),
                    description="Unit for bandwidth rates ('bit' or 'byte')"),
        OptionValue('unitprefix.bandwidth', default='metric', options=('metric', 'binary'),
                    description=("Unit prefix for bandwidth rates ('metric' or 'binary')")),

        OptionValue('unit.size', default='byte', options=('bit', 'byte'),
                    description="Unit for sizes ('bit' or 'byte')"),
        OptionValue('unitprefix.size', default='metric', options=('metric', 'binary'),
                    description=("Unit prefix for sizes ('metric' or 'binary')")),
    )


from .settings_server import (RateLimitSrvValue, PathSrvValue, PathIncompleteSrvValue)

def init_server_defaults(cfg, settingsapi):
    cfg.load(
        RateLimitSrvValue('srv.limit.rate.up',
                          description='Combined upload rate limit',
                          getter=lambda: settingsapi['rate-limit-up'],
                          setter=settingsapi.set_rate_limit_up),
        RateLimitSrvValue('srv.limit.rate.down',
                          description='Combined download rate limit',
                          getter=lambda: settingsapi['rate-limit-down'],
                          setter=settingsapi.set_rate_limit_down),

        PathSrvValue('srv.path.complete',
                     description='Where to put torrent files',
                     getter=lambda: settingsapi['path-complete'],
                     setter=settingsapi.set_path_complete),

        PathIncompleteSrvValue('srv.path.incomplete',
                     description='Where to put incomplete torrent files',
                     getter=lambda: settingsapi['path-incomplete'],
                     setter=settingsapi.set_path_incomplete),
    )


DEFAULT_KEYMAP = (
    {'context': None, 'key': 'h', 'action': '<left>'},
    {'context': None, 'key': 'j', 'action': '<down>'},
    {'context': None, 'key': 'k', 'action': '<up>'},
    {'context': None, 'key': 'l', 'action': '<right>'},
    {'context': None, 'key': 'ctrl-n', 'action': '<down>'},
    {'context': None, 'key': 'ctrl-p', 'action': '<up>'},
    {'context': None, 'key': 'ctrl-f', 'action': '<pgdn>'},
    {'context': None, 'key': 'ctrl-b', 'action': '<pgup>'},

    {'context': 'main', 'key': 'q', 'action': 'quit'},
    {'context': 'main', 'key': '?', 'action': 'tab help keymap'},
    {'context': 'main', 'key': ':', 'action': 'tui show cli'},
    {'context': 'main', 'key': 'ctrl-l', 'action': 'clearlog'},
    {'context': 'main', 'key': 'meta-L', 'action': 'tui toggle log'},
    {'context': 'main', 'key': 'meta-M', 'action': 'tui toggle main'},
    {'context': 'main', 'key': 'meta-T', 'action': 'tui toggle topbar'},
    {'context': 'main', 'key': 'meta-B', 'action': 'tui toggle bottombar'},

    {'context': 'main', 'key': 'shift-up',    'action': 'set srv.limit.rate.down +=100kB'},
    {'context': 'main', 'key': 'shift-down',  'action': 'set srv.limit.rate.down -=100kB'},
    {'context': 'main', 'key': 'shift-right', 'action': 'set srv.limit.rate.up +=100kB'},
    {'context': 'main', 'key': 'shift-left',  'action': 'set srv.limit.rate.up -=100kB'},

    {'context': 'tabs', 'key': 't', 'action': 'tab'},
    {'context': 'tabs', 'key': 'd', 'action': 'tab --close'},
    {'context': 'tabs', 'key': 'meta-1', 'action': 'tab --focus 1'},
    {'context': 'tabs', 'key': 'meta-2', 'action': 'tab --focus 2'},
    {'context': 'tabs', 'key': 'meta-3', 'action': 'tab --focus 3'},
    {'context': 'tabs', 'key': 'meta-4', 'action': 'tab --focus 4'},
    {'context': 'tabs', 'key': 'meta-5', 'action': 'tab --focus 5'},
    {'context': 'tabs', 'key': 'meta-6', 'action': 'tab --focus 6'},
    {'context': 'tabs', 'key': 'meta-7', 'action': 'tab --focus 7'},
    {'context': 'tabs', 'key': 'meta-8', 'action': 'tab --focus 8'},
    {'context': 'tabs', 'key': 'meta-9', 'action': 'tab --focus 9'},
    {'context': 'tabs', 'key': 'meta-0', 'action': 'tab --focus 10'},

    {'context': 'tabs', 'key': 'a', 'action': 'ls active'},
    {'context': 'tabs', 'key': 'A', 'action': 'tab ls active'},
    {'context': 'tabs', 'key': 'i', 'action': 'ls isolated'},
    {'context': 'tabs', 'key': 'I', 'action': 'tab ls isolated'},
    {'context': 'tabs', 'key': 's', 'action': 'ls stopped'},
    {'context': 'tabs', 'key': 'S', 'action': 'tab ls stopped'},
    {'context': 'tabs', 'key': 'z', 'action': 'ls'},
    {'context': 'tabs', 'key': 'Z', 'action': 'tab ls'},

    {'context': 'torrent', 'key': 'v', 'action': 'verify'},
    {'context': 'torrent', 'key': 'p', 'action': 'stop --toggle'},
    {'context': 'torrent', 'key': 'P', 'action': 'start --toggle --force'},
    {'context': 'torrent', 'key': 'delete', 'action': 'remove'},
    {'context': 'torrent', 'key': 'shift-delete', 'action': 'remove --delete-files'},
    {'context': 'torrent', 'key': 'enter', 'action': 'tab filelist'},

    {'context': 'file', 'key': '+', 'action': 'priority high'},
    {'context': 'file', 'key': '=', 'action': 'priority normal'},
    {'context': 'file', 'key': '-', 'action': 'priority low'},
    {'context': 'file', 'key': '0', 'action': 'priority shun'},
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Base classes for torrent commands"""

from ...logging import make_logger
log = make_logger(__name__)

import asyncio

from .. import (InitCommand, ExpectedResource)
from . import mixin


class AddTorrentsCmdbase(metaclass=InitCommand):
    name = 'add'
    aliases = ('download','get')
    provides = set()
    category = 'torrent'
    description = 'Download torrents'
    usage = ('add <TORRENT> <TORRENT> <TORRENT> ... [<OPTIONS>]',)
    examples = ('add 72d7a3179da3de7a76b98f3782c31843e3f818ee',
                'add --stopped http://example.org/something.torrent')
    argspecs = (
        { 'names': ('TORRENT',), 'nargs': '+',
          'description': 'Link or path to torrent file, magnet link or hash' },
        { 'names': ('--stopped','-s'), 'action': 'store_true',
          'description': 'Do not start downloading the added torrent' },
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT, stopped):
        success = True
        force_torrentlist_update = False
        for source in TORRENT:
            response = await self.make_request(self.srvapi.torrent.add(source, stopped=stopped))
            success = success and response.success
            force_torrentlist_update = force_torrentlist_update or success

        # Update torrentlist AFTER all 'add' requests
        if force_torrentlist_update and hasattr(self, 'polling_frenzy'):
            self.polling_frenzy()
        return success


class ListTorrentsCmdbase(mixin.get_torrent_sorter, mixin.get_tlist_columns,
                          metaclass=InitCommand):
    name = 'list'
    aliases = ('ls',)
    provides = set()
    category = 'torrent'
    description = 'List torrents'
    usage = ('list [<OPTIONS>]',
             'list [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('ls active',
                'ls !active',
                'ls seeds<10',
                'ls active&tracker~example.org',
                'ls active|idle&tracker~example')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`)'},

        { 'names': ('--sort', '-s'),
          'default_description': "current value of 'tlist.sort' setting",
          'description': ('Comma-separated list of sort orders '
                          "(see 'sort' command for available sort orders)") },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'tlist.columns' setting",
          'description': ('Comma-separated list of column names '
                          "(see 'help tlist.columns' for available columns)") },
    )
    more_sections = {
        'SCRIPTING': (
            ("If invoked as a command line argument and the output does not "
             "go to a TTY (i.e. the terminal size can't be determined), "
             "the output is optimized for scripting.  Numbers are "
             "unformatted, columns are separated by '|' and headers are "
             "not printed."),
            "",
            ("To enforce human-readable, formatted output, set the environment"
             "variables COLUMNS and LINES."),
            "",
            "\t$ \tCOLUMNS=80 LINES=24 {APPNAME} ls | less -R"
        ),
    }

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, sort, columns):
        sort = self.cfg['tlist.sort'].value if sort is None else sort
        columns = self.cfg['tlist.columns'].value if columns is None else columns
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=True,
                                           discover_torrent=False)
            sort = self.get_torrent_sorter(sort)
            columns = self.get_tlist_columns(columns)
        except ValueError as e:
            log.error(e)
            return False
        else:
            log.debug('Listing %s torrents sorted by %s', tfilter, sort)
            if asyncio.iscoroutinefunction(self.make_tlist):
                return await self.make_tlist(tfilter, sort, columns)
            else:
                return self.make_tlist(tfilter, sort, columns)


class ListFilesCmdbase(mixin.get_flist_columns, metaclass=InitCommand):
    name = 'filelist'
    aliases = ('fls', 'lsf')
    provides = set()
    category = 'torrent'
    description = 'List torrent files'
    usage = ('filelist [<OPTIONS>]',
             'filelist [<OPTIONS>] [<TORRENT FILTER>] [<FILE FILTER>]')
    examples = ('filelist',
                "filelist A.Torrent.with.Files",
                "filelist A.Torrent.with.Files priority=low")
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},

        { 'names': ('FILE FILTER',), 'nargs': '?',
          'description': 'Filter expression (see `help filter`)' },

        { 'names': ('--columns', '-c'),
          'default_description': "current value of 'flist.columns' setting",
          'description': ('Comma-separated list of column names '
                          "(see 'help flist.columns' for available columns)") },
    )
    more_sections = {'SCRIPTING': ListTorrentsCmdbase.more_sections['SCRIPTING']}

    cfg = ExpectedResource

    async def run(self, TORRENT_FILTER, FILE_FILTER, columns):
        columns = self.cfg['flist.columns'].value if columns is None else columns
        try:
            columns = self.get_flist_columns(columns)
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
            ffilter = self.select_files(FILE_FILTER,
                                        allow_no_filter=True,
                                        discover_file=False)
        except ValueError as e:
            log.error(e)
            return False

        log.debug('Listing %s files of %s torrents', ffilter, tfilter)

        if asyncio.iscoroutinefunction(self.make_flist):
            return await self.make_flist(tfilter, ffilter, columns)
        else:
            return self.make_flist(tfilter, ffilter, columns)


class PriorityCmdbase(metaclass=InitCommand):
    name = 'priority'
    aliases = ('prio',)
    provides = set()
    category = 'torrent'
    description = 'Change download priority of files'
    usage = ('priority <PRIORITY>',
             'priority <PRIORITY> [<TORRENT FILTER>] [<FILE FILTER>]')
    examples = ('priority low',
                'priority high "some torrent" size>12M')
    argspecs = (
        { 'names': ('PRIORITY',),
          'description': ("File priority; must be low/normal/high/shun, "
                          "l/n/h/s or -/./+/0")},
        {'names': ('TORRENT FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
        { 'names': ('FILE FILTER',), 'nargs': '?',
         'description': 'Filter expression (see `help filter`) or focused file in the TUI'},
    )

    srvapi = ExpectedResource

    _PRIORITY = {'l': 'low',    '-': 'low',    'low': 'low',
                 'n': 'normal', '.': 'normal', 'normal': 'normal',
                 'h': 'high',   '+': 'high'  , 'high': 'high',
                 's': 'shun',   '0': 'shun',   'shun': 'shun'}

    async def run(self, PRIORITY, TORRENT_FILTER, FILE_FILTER):
        try:
            priority = self._PRIORITY[PRIORITY.lower()]
        except KeyError:
            log.error('Invalid priority: {!r}'.format(PRIORITY))
            return False

        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
            ffilter = self.select_files(FILE_FILTER,
                                        allow_no_filter=True,
                                        discover_file=True)
        except ValueError as e:
            log.error(e)
            return False

        log.debug('Setting file download priority to %s for %s files of %s torrents',
                  priority, ffilter, tfilter)

        if not isinstance(tfilter, tuple):
            # tfilter must be TorrentFilter instance, which means the user
            # specified a filter and will be informed about the matches.
            if isinstance(ffilter, tuple):
                # The user did specify a torrent filter but not a file filter,
                # so select_files() may have returned a focused file.  But we
                # assume that the user meant all files of the matching
                # torrents, otherwise they wouldn't have given a torrent
                # filter.
                ffilter = None

            msg = 'New download priority of %s files in %s torrents: %s' % (
                'all' if ffilter is None else ffilter, tfilter, priority)
            log.info(msg)
            quiet = False
        else:
            # We're operating on the focused file and success is indiciated by
            # the updated file list, so no info message necessary.
            quiet = True

        response = await self.make_request(
            self.srvapi.torrent.file_priority(priority, tfilter, ffilter),
            polling_frenzy=True, quiet=quiet)
        return response.success


class RemoveTorrentsCmdbase(metaclass=InitCommand):
    name = 'remove'
    aliases = ('rm', 'delete')
    provides = set()
    category = 'torrent'
    description = 'Remove torrents'
    usage = ('remove [<OPTIONS>]',
             'remove [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('remove',
                'remove "some torrent" another\ torrent and_this_torrent',
                'remove -d "unwanted torrent"')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},

        { 'names': ('--delete-files','-d'), 'action': 'store_true',
          'description': 'Delete any downloaded files' },
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, delete_files):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            response = await self.make_request(
                self.srvapi.torrent.remove(tfilter, delete=delete_files),
                polling_frenzy=True)
            return response.success


# Argument definitions that are shared between commands
ARGSPEC_TOGGLE = {
    'names': ('--toggle','-t'), 'action': 'store_true',
    'description': ('Start TORRENT if stopped and vice versa')
}

class StartTorrentsCmdbase(metaclass=InitCommand):
    name = 'start'
    aliases = ()
    provides = set()
    category = 'torrent'
    description = 'Start downloading torrents'
    usage = ('start [<OPTIONS>]',
             'start [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('start',
                "start 'night of the living dead' Metropolis",
                'start ubuntu --force')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
        { 'names': ('--force','-f'), 'action': 'store_true',
          'description': 'Ignore download queue' },
        ARGSPEC_TOGGLE,
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, toggle, force):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            if toggle:
                response = await self.make_request(
                    self.srvapi.torrent.toggle_stopped(tfilter, force=force),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    self.srvapi.torrent.start(tfilter, force=force),
                    polling_frenzy=True)
            return response.success


class StopTorrentsCmdbase(metaclass=InitCommand):
    name = 'stop'
    aliases = ('pause',)
    provides = set()
    category = 'torrent'
    description = 'Stop downloading torrents'
    usage = ('stop [<OPTIONS>]',
             'stop [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('stop',
                'stop "night of the living dead" idle',
                'stop --toggle ubuntu')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
        ARGSPEC_TOGGLE,
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER, toggle):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            if toggle:
                response = await self.make_request(
                    self.srvapi.torrent.toggle_stopped(tfilter),
                    polling_frenzy=True)
            else:
                response = await self.make_request(
                    self.srvapi.torrent.stop(tfilter),
                    polling_frenzy=True)
            return response.success


class VerifyTorrentsCmdbase(metaclass=InitCommand):
    name = 'verify'
    aliases = ('check',)
    provides = set()
    category = 'torrent'
    description = 'Verify downloaded torrent data'
    usage = ('verify [<OPTIONS>]',
             'verify [<OPTIONS>] [<TORRENT FILTER> <TORRENT FILTER> ...]')
    examples = ('verify',
                'verify debian')
    argspecs = (
        {'names': ('TORRENT FILTER',), 'nargs': '*',
         'description': 'Filter expression (see `help filter`) or focused torrent in the TUI'},
    )

    srvapi = ExpectedResource

    async def run(self, TORRENT_FILTER):
        try:
            tfilter = self.select_torrents(TORRENT_FILTER,
                                           allow_no_filter=False,
                                           discover_torrent=True)
        except ValueError as e:
            log.error(e)
            return False
        else:
            response = await self.make_request(self.srvapi.torrent.verify(tfilter),
                                               polling_frenzy=False)
            return response.success



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

from ..settings import is_server_setting


class CommandCache():
    """Store rc/CLI commands"""

    def __init__(self, cmdmgr):
        self._cmdmgr = cmdmgr
        self.rccmds = []
        self.clicmds = []

    def add_rccmds(self, *cmds):
        self.rccmds.extend(*cmds)
        log.debug('Cached rc commands: %r', self.rccmds)

    def add_clicmds(self, *cmds):
        self.clicmds.extend(*cmds)
        log.debug('Cached CLI commands: %r', self.clicmds)

    def select(self, source='ANY', interface='ANY'):
        """Return commands from `source` that provide `interface`

        source: 'cli' (command line) or 'rc' (rc file)
        interface: Provided interface ('cli' or 'tui')
        """
        if source == 'cli':
            cmdlines = self.clicmds
        elif source == 'rc':
            cmdlines = self.rccmds
        elif source == 'ANY':
            cmdlines = self.clicmds + self.rccmds
        else:
            raise ValueError('Invalid source: {!r}'.format(source))

        result = []
        for cmdline in cmdlines:
            cmdname = cmdline[0]
            cmdcls = self._cmdmgr.get_cmdcls(cmdname, interface=interface)
            if cmdcls is not None:
                result.append(cmdline)

        log.debug('Selected %s commands for %s interface: %s',
                  source, interface, result)
        return result

    def guess_ui(self, cmds):
        """Guess user interface (TUI or CLI) based on CLI commands"""
        if not cmds:
            return 'tui'

        log.debug('Guessing if the TUI is wanted:')
        guess = 'tui'
        for cmdline in cmds:
            cmdname = cmdline[0]
            debugmsg = '  %s: ' % cmdname

            tuicmd = self._cmdmgr.get_cmdcls(cmdname, interface='tui')
            clicmd = self._cmdmgr.get_cmdcls(cmdname, interface='cli')
            if tuicmd is None is clicmd:
                # Unknown command - maybe command is not loaded yet?
                debugmsg += 'unknown command - no guess'
                continue

            # Does command provides only one interface?
            elif tuicmd is None:
                debugmsg += 'no support for TUI - guessing CLI'
                guess = 'cli'   # Command NEEDS CLI
            elif clicmd is None:
                debugmsg += 'no support for CLI - guessing TUI'
                guess = 'tui'   # Command NEEDS TUI

            elif tuicmd.category == clicmd.category == 'torrent':
                # Torrent commands (start, stop, list, ...) inhibit the TUI
                debugmsg += 'torrent command - guessing CLI'
                guess = 'cli'

            # Some 'set' commands should enforce the tui or cli, other 'set'
            # commands shouldn't care.
            elif cmdline[0] == 'set' and len(cmdline) >= 3 or \
                 cmdline[0] == 'reset' and len(cmdline) >= 2:
                setting = cmdline[1]
                if setting.startswith('tui.'):
                    debugmsg += '{!r} starts with \'tui.\' - guessing TUI'.format(setting)
                    guess = 'tui'
                elif is_server_setting(setting):
                    debugmsg += 'is_server_setting({!r}) == True - guessing CLI'.format(setting)
                    guess = 'cli'

            elif clicmd is not None:
                debugmsg += 'CLI supported - guessing CLI'
                guess = 'cli'

            else:
                debugmsg += 'no guess'

        log.debug(debugmsg)
        log.debug('Guessed wanted UI: %s', guess)
        return guess
